from functools import wraps
from typing import Any, Awaitable, Callable, ParamSpec, Sequence, TypeVar

from .cache import Cache

P = ParamSpec("P")
T = TypeVar("T")


def _extend_posargs(sig: list, posargs: list[int], *args) -> None:
    for i in posargs:
        val = args[i]

        hashed = hash(val)

        sig.append(hashed)


def _extend_kwargs(sig: list, _kwargs: list[str], **kwargs) -> None:
    for name in _kwargs:
        val = kwargs[name]

        hashed = hash(val)

        sig.append(hashed)


def _get_sig(
    include_posargs: list[int] = None, include_kwargs: list[str] = None, *args, **kwargs
) -> Sequence[int]:
    signature = []

    if include_posargs:
        _extend_posargs(signature, include_posargs, *args)

    if include_kwargs:
        _extend_kwargs(signature, include_kwargs, **kwargs)

    return tuple(signature)


def cached(
    timeout: float = None,
    include_posargs: list[int] = None,
    include_kwargs: list[str] = None,
    cache: Cache = None,
) -> Callable[[Callable[P, T]], Callable[P, T]]:
    _cache: Cache[Sequence[int], Any] = cache or Cache(timeout=timeout)

    def decorator(func: Callable[P, T]) -> Callable[P, T]:
        @wraps(func)
        def wrapper(*args, **kwargs) -> T:
            sig = _get_sig(include_posargs, include_kwargs, *args, **kwargs)

            if value := _cache.get(sig):
                return value

            retval = func(*args, **kwargs)

            _cache[sig] = retval

            return retval

        return wrapper

    return decorator


def acached(
    timeout: float = None,
    include_posargs: list[int] = None,
    include_kwargs: list[str] = None,
    cache: Cache = None,
) -> Callable[[Callable[P, Awaitable[T]]], Callable[P, Awaitable[T]]]:
    _cache: Cache[Sequence[int], Any] = cache or Cache(timeout=timeout)

    def decorator(func: Callable[P, Awaitable[T]]) -> Callable[P, Awaitable[T]]:
        @wraps(func)
        async def awrapper(*args, **kwargs) -> T:
            sig = _get_sig(include_posargs, include_kwargs, *args, **kwargs)

            if value := _cache.get(sig):
                return value

            retval: T = await func(*args, **kwargs)

            _cache[sig] = retval

            return retval

        return awrapper

    return decorator



from time import sleep, time
from typing import Generic, TypeVar

KT = TypeVar("KT")
VT = TypeVar("VT")
T = TypeVar("T")


class _Expirable(Generic[VT]):
    """
    Base class for objects that can expire.
    """

    def __init__(self, value: VT, timeout: float = None) -> None:
        self.value = value

        self._expires = timeout + time() if timeout else None

    @property
    def expired(self) -> bool:
        if not self._expires:
            return False
        return self._expires < time()


class Cache(Generic[KT, VT]):
    def __init__(self, values: dict[KT, VT] = None, timeout: float = None) -> None:
        self._items: dict[KT, _Expirable[VT]] = (
            {key: _Expirable(values, timeout) for key, values in values.items()}
            if values
            else {}
        )
        self._timeout = timeout

    def __getitem__(self, key: KT) -> VT:
        item = self._items[key]

        if item.expired:
            raise KeyError(key)

        return item.value

    def __setitem__(self, key: KT, value: VT) -> None:
        self._items[key] = _Expirable(value, self._timeout)

    def get(self, key: KT, default: T | None = None) -> VT | T | None:
        try:
            return self[key]
        except KeyError:
            return default

    def __contains__(self, key: KT) -> bool:
        return key in self._items and not self._items[key].expired



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cachingutils']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'cachingutils',
    'version': '1.0.0',
    'description': 'Utilities to make caching data easier',
    'long_description': '# cachingutils\n\nUtilities to make caching data easier\n\n## Examples\n\nBasic caching:\n\n```py\nfrom cachingutils import cached\n\n\n@cached()\ndef fib(n: int) -> int:\n    if n < 2:\n        return n\n    return fib(n - 1) + fib(n - 2)\n\nprint(fib(100))  # 633825300114114700748351602688\n```\n\nCaching with your own cache object:\n\n```py\nfrom cachingutils import Cache, cached\n\n\nmy_cache = Cache()\n\n@cached(cache=my_cache)\ndef fib(n: int) -> int:\n    if n < 2:\n        return n\n    return fib(n - 1) + fib(n - 2)\n\nprint(fib(100))  # 633825300114114700748351602688\n```\n\nAsync caching:\n\n```py\nfrom asyncio import run\n\nfrom cachingutils import acached\n\n\n@acached()\nasync def fib(n: int) -> int:\n    if n < 2:\n        return n\n    return await fib(n - 1) + await fib(n - 2)\n\nprint(run(fib(100)))  # 633825300114114700748351602688\n```\n\nCaching specific positional args:\n\n```py\nfrom cachingutils import cached\n\n\n@cached(include_posargs=[0])\nasync def add(a: int, b: int) -> int:\n    return a + b\n\nprint(add(1, 2))  # 3\nprint(add(2, 2))  # 3\nprint(add(2, 3))  # 5\n```\n\nCaching specific keyword args:\n\n```py\nfrom cachingutils import cached\n\n\n@cached(include_posargs=[0], include_kwargs=[\'c\'])\ndef add(a: int, b: int, *, c: int) -> int:\n    return a + b\n\nprint(add(1, 2, c=3))  # 3\nprint(add(2, 2, c=3))  # 4\nprint(add(2, 3, c=3))  # 4\n```\n\nCaching with a timeout:\n\n```py\nfrom time import sleep\n\nfrom cachingutils import cached\n\n\n@cached(timeout=1, include_posargs=[0])\ndef add(a: int, b: int) -> int:\n    return a + b\n\nprint(add(1, 2))  # 3\nprint(add(1, 3))  # 3\nsleep(2)\nprint(add(1, 3))  # 4\n```\n\nUsing a raw `Cache` object:\n\n```py\nfrom time import sleep\n\nfrom cachingutils import Cache\n\n\nmy_cache: Cache[str, int] = Cache(timeout=5)\n\nmy_cache["abc"] = 123\n\nprint(my_cache["abc"])  # 123\n\nsleep(6)\n\nprint(my_cache["abc"])  # KeyError: \'abc\'\n```\n',
    'author': 'vcokltfre',
    'author_email': 'vcokltfre@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/vcokltfre/cachingutils',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
